import sys
sys.path.append('src')

# import pytest # required for pytest.raises

from mvin.functions.excel_lib import excel_search
from mvin import TokenString, TokenNumber, TokenError, TokenErrorTypes

def test_excel_search_found():
    find_text = TokenString("world")
    within_text = TokenString("hello world")
    start_num = TokenNumber(1)
    result = excel_search(find_text, within_text, start_num)
    assert isinstance(result, TokenNumber)
    assert result.value == 7

def test_excel_search_not_found():
    find_text = TokenString("world")
    within_text = TokenString("hello")
    start_num = TokenNumber(1)
    result = excel_search(find_text, within_text, start_num)
    assert isinstance(result, TokenError)
    assert result.value == TokenErrorTypes.VALUE.value

def test_excel_search_with_start_num():
    find_text = TokenString("world")
    within_text = TokenString("hello world world")
    start_num = TokenNumber(8)
    result = excel_search(find_text, within_text, start_num)
    assert isinstance(result, TokenNumber)
    assert result.value == 13

def test_excel_search_with_negative_start_num():
    find_text = TokenString("world")
    within_text = TokenString("hello world world")
    start_num = TokenNumber(-8)
    result = excel_search(find_text, within_text, start_num)
    assert isinstance(result, TokenError)
    assert result.value == TokenErrorTypes.VALUE.value
    assert result.message == "Expected integer >= 1 for start_num argument, but found: -8"


def test_excel_search_invalid_start_num():
    find_text = TokenString("world")
    within_text = TokenString("hello world")
    start_num = TokenString("invalid")
    result = excel_search(find_text, within_text, start_num)
    assert isinstance(result, TokenError)
    assert result.value == TokenErrorTypes.VALUE.value

def test_excel_search_error_in_start_num():
    find_text = TokenString("world")
    within_text = TokenString("hello world")
    start_num = TokenError(TokenErrorTypes.VALUE, "Error")
    result = excel_search(find_text, within_text, start_num)
    assert isinstance(result, TokenError)
    assert result.value == TokenErrorTypes.VALUE.value



import logging
from typing import Callable, Dict, List, Tuple, Union
from mvin import Token, TokenBool, TokenError, TokenErrorTypes, TokenNumber, TokenString


def excel_not(token: Token) -> Token:
    if token and token.type == "OPERAND":
        if token.subtype == "LOGICAL":
            return TokenBool(not token.value)
        elif token.subtype == "NUMBER":
            return TokenBool(not (token.value != 0))
    return TokenError(
        TokenErrorTypes.VALUE,
        f"Expected boolean or number and found {token}",
    )


def excel_iserror(token: Token) -> Token:
    return TokenBool(token and token.subtype == "ERROR")


def excel_search(
    find_text: Union[Token, None],
    within_text: Union[Token, None],
    start_num: Union[Token, None],
) -> Token:
    logging.debug(
        f"excel_lib.excel_search: calling with (find_text= {find_text}, within_text={within_text}, start_num= {start_num} )"
    )
    if not within_text:
        return TokenError(TokenErrorTypes.VALUE, "Argument within_text cannot be None")
    if within_text.type != "OPERAND":
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Expected value for within_text argument, but found: {within_text}",
        )
    if within_text.subtype == "ERROR":
        return within_text

    within_text_value = ""
    if within_text:
        if within_text.subtype == "ERROR":
            return within_text
        elif within_text.subtype == "TEXT":
            within_text_value = within_text.value
        elif within_text.subtype == "NUMBER":
            within_text_value = str(within_text.value)
        elif within_text.subtype == "LOGICAL":
            within_text_value = "TRUE" if within_text.value else "FALSE"
        else:
            return TokenError(
                TokenErrorTypes.VALUE,
                f"Unsupported value type for within_text argument: {within_text}",
            )

    find_text_value = ""
    if find_text:
        if find_text.subtype == "ERROR":
            return find_text
        elif find_text.subtype == "TEXT":
            find_text_value = find_text.value
        elif find_text.subtype == "NUMBER":
            find_text_value = str(find_text.value)
        elif find_text.subtype == "LOGICAL":
            find_text_value = "TRUE" if find_text.value else "FALSE"
        else:
            return TokenError(
                TokenErrorTypes.VALUE,
                f"Unsupported value type for find_text argument: {find_text}",
            )

    start_num_value = 0
    if start_num:
        if start_num.subtype == "ERROR":
            return start_num
        elif start_num.subtype == "NUMBER":
            start_num_value = start_num.value - 1

            if start_num_value < 0:
                return TokenError(
                    TokenErrorTypes.VALUE,
                    f"Expected integer >= 1 for start_num argument, but found: {start_num.value}",
                )
        else:
            return TokenError(
                TokenErrorTypes.VALUE,
                f"Expected integer valur for start_num argument, but found: {start_num}",
            )

    found_index = within_text_value.find(find_text_value, start_num_value)
    if found_index >= 0:
        return TokenNumber(found_index + 1)
    else:
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Text `{find_text_value}` not found in text `{within_text_value} (starting at: {start_num_value + 1})`",
        )


DEFAULT_FUNCTIONS: Dict[str, Tuple[Union[List, None], Callable]] = {
    "NOT(": (
        [
            None
        ],  # default argument list (if None is in the list, that argument is not optional)
        excel_not,
    ),
    "ISERROR(": (
        [
            None
        ],  # default argument list (if None is in the list, that argument is not optional)
        excel_iserror,
    ),
    "SEARCH(": (
        [
            TokenString(""),  # find_text
            None,  # within_text
            TokenNumber(1),  # start_num --> default: 1
        ],  # default argument list (if None is in the list, that argument is not optional)
        excel_search,
    ),
}



import logging
import typing
from collections import deque
from types import MappingProxyType
from typing import Any, Callable, Dict, List, Mapping, Sequence, Set, Tuple, Union

import mvin.excel_ops as _  # noqa

from mvin import REGISTERED_OPS, Token, TokenError, TokenErrorTypes, TokenFunc
from mvin.functions.excel_lib import DEFAULT_FUNCTIONS

# Operator precedence and associativity
OPERATORS: Mapping[str, Tuple[Union[int, float], str]] = MappingProxyType(
    {
        "&": (0.5, "L"),  # String concatenation
        "=": (0, "L"),
        "==": (0, "L"),
        "<>": (0, "L"),
        "!=": (0, "L"),
        "<": (0, "L"),
        ">": (0, "L"),
        "<=": (0, "L"),
        ">=": (0, "L"),
        "+": (1, "L"),
        "-": (1, "L"),
        "*": (2, "L"),
        "/": (2, "L"),
        "^": (3, "R"),  # Exponentiation is right-associative
    }
)


def get_interpreter(
    tokens: Sequence[Token],  # enumerable
    proposed_functions: Dict[
        str, Tuple[Union[List, None], Callable]
    ] = DEFAULT_FUNCTIONS,
    registered_ops: Dict[str, Callable[[Token, Token], Token]] = REGISTERED_OPS,
) -> Union[Callable[[Dict[str, Any]], Any], None]:
    if isinstance(tokens, Sequence):
        ops = MappingProxyType(registered_ops)
        functions = MappingProxyType(proposed_functions)

        def infix_to_rpn(
            tokens: Sequence[Token],  # tokens from the tokenizer
            functions: Mapping[str, Tuple[Union[List, None], Callable]],
        ) -> Tuple[List[Union[Token, int, None]], Set[str]]:
            """
            Convert an infix expression to Reverse Polish Notation (RPN).

                Args:
                    tokens (list): A list of tokens representing the infix expression.

                Returns:
                    list: A list of tokens in Reverse Polish Notation (RPN).

                Raises:
                    SyntaxError: If there are syntax errors in the infix expression.

                The function processes the input tokens and converts them to RPN format.
                It handles operators, operands, parentheses, and functions, ensuring proper
                syntax and operator precedence. The function uses two stacks to manage
                operators and argument counts, and includes detailed error handling to
                provide informative messages for various syntax issues.
            """
            output: List[Union[Token, int, None]] = []
            op_stack: deque[Token] = deque()
            arg_stack: deque[int] = (
                deque()
            )  # Keeps track of argument counts for nested function calls
            open_parens = 0

            inputs: Set[str] = set()

            for i, token in enumerate(
                # Filter whitespace to simply logic
                [x for x in tokens if x is None or x.type != "WHITE-SPACE"]
            ):
                logging.debug(
                    f"--------\noutput: {output}\nop_stack: {op_stack}\narg_count: {arg_stack}\ntoken: {token}"
                )
                if token is None:
                    raise SyntaxError(f"Unexpected None value found at position {i}.")
                if token.type == "OPERAND":
                    output.append(token)

                    if token.subtype == "RANGE":
                        inputs.add(token.value)

                elif token.type == "FUNC" and token.subtype == "OPEN":
                    func_name = token.value
                    if func_name not in functions:
                        raise SyntaxError(
                            f"Unsupported function `{token.value}` at position {i}."
                        )
                    # op_stack.append(func_name)
                    op_stack.append(token)
                    arg_stack.append(
                        1
                    )  # Start with one argument for this instance of the function

                    op_stack.append(TokenFunc("("))
                    open_parens += 1

                elif token.type == "OPERATOR-INFIX":
                    if i == 0 or (
                        tokens[i - 1].type == "OPERATOR-INFIX"
                        or tokens[i - 1].value == "("
                    ):
                        raise SyntaxError(
                            f"Unexpected operator `{token}` at position {i}."
                        )
                    token_value = token.value
                    while (
                        op_stack
                        and op_stack[-1].value in OPERATORS
                        and (
                            (
                                OPERATORS[token_value][1] == "L"
                                and OPERATORS[token_value][0]
                                <= OPERATORS[op_stack[-1].value][0]
                            )
                            or (
                                OPERATORS[token_value][1] == "R"
                                and OPERATORS[token_value][0]
                                < OPERATORS[op_stack[-1].value][0]
                            )
                        )
                    ):
                        output.append(op_stack.pop())  # TODO Improve test coverage
                    op_stack.append(token)

                elif token.value == "(":  # Left parenthesis
                    if i > 0 and tokens[i - 1].type == "OPERAND":
                        raise SyntaxError(
                            f"Missing operator before '(' at position {i}."
                        )
                    op_stack.append(token)
                    open_parens += 1

                elif token.value == ")":  # Right parenthesis
                    open_parens -= 1
                    if open_parens < 0:
                        raise SyntaxError(
                            f"Unexpected `)` at position {i} (too many closing parentheses)."
                        )

                    # Ensure trailing empty argument handling
                    last_token = tokens[i - 1] if i > 0 else None
                    logging.debug(f"last_token: {last_token}")
                    if last_token:
                        if last_token.type == "FUNC" and last_token.subtype == "OPEN":
                            arg_stack[-1] = 0

                        elif last_token.value == ",":
                            output.append(None)  # Handle missing last argument

                    while op_stack and op_stack[-1].value != "(":
                        output.append(op_stack.pop())

                    if not op_stack:
                        raise SyntaxError(f"Unmatched `)` at position {i}.")
                    op_stack.pop()  # Remove '('

                    if op_stack and op_stack[-1].type == "FUNC":
                        #!!! DESIGN DECISION
                        #
                        # Support closing parenthesis, even when is not of type='FUNC' --> display warning
                        if token.type != "FUNC":
                            logging.warning(
                                f"Expected token of type:`FUNC` and subtype:`CLOSE`, but found token of type:`{token.type}`"
                            )
                        # if op_stack and token.type == "FUNC" and token.subtype == "CLOSE":
                        # func_name = op_stack.pop()
                        func = op_stack.pop()
                        func_name = func.value
                        arg_count = arg_stack.pop()  # Use dynamic argument count

                        #!!! DESIGN DECISION
                        #
                        # arg_count full validation: We are filling the missing part of the arguments
                        #
                        required_args, _ = functions[func_name]
                        if required_args is not None:
                            required_args_count = len(required_args)

                            if required_args_count < arg_count:
                                raise SyntaxError(
                                    f"Function `{func_name}` expects {required_args_count} arguments but got {arg_count}."
                                )

                            # copy defaults to stack if not None
                            # print(
                            #     f" --> arg_count:{arg_count} required_args_count:{required_args_count}"
                            # )
                            arg_index = arg_count
                            while arg_index < required_args_count:
                                default_arg = required_args[arg_index]
                                # print(
                                #     f"   --> arg_index:{arg_index} default_arg: {default_arg}"
                                # )
                                if default_arg is not None:
                                    output.append(default_arg)
                                else:
                                    raise SyntaxError(
                                        f"Missing required argument at {arg_index} for function `{func_name}`"
                                    )
                                arg_index += 1
                            arg_count = required_args_count

                        #!!! DESIGN DECISION
                        #
                        # There are 2 ways to go to correct the arg_count bug:
                        #  - Use the stack to store the arg_count:
                        #    * append the arg_count first as int (and in the execute part, read it from stack)
                        #    * append the TokenFunc
                        #
                        #    > This option pollutes the stack by mixing types
                        #    > There is a possibility to have more dynamic variable arguments calling by putting
                        #      the value in the stack
                        #
                        #  - Use the output to store the arg_count and read it directly when execute
                        #    * appent the TokenFunc
                        #    * append the arg_count as int (and read it just after the TokenFunc, and it will never reach the Token evaluation)
                        #
                        #    > This options requires an additional index to track the rpn_tokens' position
                        #
                        #  I preferred the non pullution version (the second one)

                        # output.append(TokenFunc(func_name))
                        output.append(func)
                        output.append(arg_count)

                elif (
                    token.type == "SEP" and token.subtype == "ARG"
                ):  # token.value == ',':
                    if tokens[i - 1].subtype == "OPEN" or tokens[i - 1].value == ",":
                        output.append(None)  # Handle missing argument

                    while op_stack and op_stack[-1].value != "(":
                        output.append(op_stack.pop())

                    if arg_stack:
                        arg_stack[-1] += (
                            1  # Increase count for the current function call
                        )

                # elif token.type == "WHITE-SPACE":
                #     pass

                else:
                    raise SyntaxError(f"Unrecognized token `{token}` at position {i}.")

            if open_parens > 0:
                raise SyntaxError("Unmatched `(` (missing closing parenthesis).")

            while op_stack:
                if op_stack[-1].value == "(":
                    raise SyntaxError("Unmatched `(` in expression.")
                output.append(op_stack.pop())

            return output, inputs

        rpn_tokens, inputs = infix_to_rpn(tokens, functions)
        # print(f"rpn_tokens: {rpn_tokens}")

        def execute_func(
            rpn_tokens: Sequence[Union[Token, int, None]], inputs_set: Set[str]
        ) -> Callable[[Dict[str, typing.Any]], Any]:
            def evaluate_rpn(inputs: Dict[str, typing.Any] = {}) -> typing.Any:
                immutable_inputs: Mapping[str, Any] = MappingProxyType(inputs)

                stack = deque()

                i = 0  # Token index for argument count retrieval

                while i < len(rpn_tokens):
                    token = rpn_tokens[i]
                    i += 1  # Move to the next token
                    # print(token)

                    if token is None:
                        # raise ValueError(f"Unexpected None at position {i}")
                        stack.append(token)
                    elif isinstance(token, int):
                        raise ValueError(
                            f"Unexpected token (int: {token}) at position {i} "
                        )
                    elif token.type == "OPERAND":
                        if token.subtype != "RANGE":
                            stack.append(token)
                        else:
                            range_value = immutable_inputs.get(token.value)
                            if range_value:
                                stack.append(range_value)
                            else:
                                raise KeyError(
                                    f"The input '{token.value}' is required but was not found in the provided inputs."
                                )
                    elif token.type == "OPERATOR-INFIX":  # in OPERATORS:
                        if len(stack) < 2:
                            raise ValueError(
                                f"Not enough values for operation '{token.value}'."
                            )
                        b = stack.pop()
                        a = stack.pop()

                        if (
                            token.value == "/"
                            and b.type == "OPERAND"
                            and (b.value == 0 or b.value == 0.0)
                        ):
                            stack.append(
                                TokenError(
                                    TokenErrorTypes.ZERO_DIV, "Division by zero."
                                )
                            )
                        else:
                            op = ops.get(token.value)
                            if op:
                                stack.append(op(a, b))
                            else:
                                raise NotImplementedError(
                                    f"Operator '{token.value}' is not implemented"
                                )
                    elif token.type == "FUNC" and token.subtype == "OPEN":
                        func_name = token.value

                        if i >= len(rpn_tokens):
                            raise ValueError(
                                f"Missing argument count for function `{func_name}`."
                            )

                        arg_count = rpn_tokens[
                            i
                        ]  # Read argument count from tokens, not stack
                        i += 1  # Move to next token after argument count

                        if isinstance(arg_count, int):
                            args = [
                                stack.pop() if stack else None for _ in range(arg_count)
                            ][::-1]
                            # print(f"`{func_name}`-> args: {args}")

                            # Get function default values
                            func_defaults, func_callable = functions[func_name]

                            if func_defaults is not None:
                                required_args_count = len(func_defaults)
                                if len(args) != required_args_count:
                                    raise ValueError(
                                        f"Function `{token}` expects {required_args_count} arguments but got {len(args)}."
                                    )
                                arg_index = 0
                                while arg_index < required_args_count:
                                    default_value = func_defaults[arg_index]
                                    # print(
                                    #     f" *-> Eval#{func_name} :: {arg_index} | {args[arg_index]} | {default_value}"
                                    # )
                                    if (
                                        default_value is None
                                        and args[arg_index] is None
                                    ):
                                        raise ValueError(
                                            f"Missing required argument at {arg_index} for function `{func_name}`"
                                        )
                                    arg_index += 1

                            func_result = func_callable(*args)
                            # print(f"`{func_name}`-> result: {func_result}")
                            if func_result:
                                stack.append(func_result)
                        else:
                            raise RuntimeError(
                                f"Expected an interger arg_count for function `{func_name}`'s arg_count, but found {arg_count}"
                            )

                if len(stack) != 1:
                    raise ValueError(
                        "Formula evaluation error: too many values remaining."
                    )

                return stack.pop().value

            evaluate_rpn.__setattr__("inputs", inputs_set)
            return evaluate_rpn

        return execute_func(rpn_tokens, inputs)
    return None



"""
This module defines various token classes and utility functions for registering and handling operations on tokens.
The valid combinations for the token classes are as follows:
    type: OPERAND
    subtype: LOGICAL | TEXT | NUMBER | ERROR | ARRAY | RANGE

Classes:
    Token: Abstract base class for all tokens.
    BaseToken: Base class implementing the Token interface.
    TokenBool: Token class for boolean values.
    TokenString: Token class for string values.
    TokenNumber: Token class for numeric values.
    TokenFunc: Token class for function names.
    TokenError: Token class for error values.
    TokenErrorTypes: Enum class for different types of token errors.

Functions:
    register_op: Decorator to register operator functions with multiple names.
    register_numeric_op: Decorator to register numeric operator functions with multiple names.

Constants:
    REGISTERED_OPS: Dictionary to store registered operations.
"""

from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Dict, Callable, Any, Tuple, Union

__version__ = "0.5.0b4"


class Token(metaclass=ABCMeta):
    """
    Abstract base class for all tokens.
    """

    @property
    @abstractmethod
    def type(self) -> str:
        """Returns the type of the token."""
        pass

    @property
    @abstractmethod
    def subtype(self) -> str:
        """Returns the subtype of the token."""
        pass

    @property
    @abstractmethod
    def value(self) -> Any:
        """Returns the value of the token."""
        pass


class BaseToken(Token):
    """
    Base class implementing the Token interface.
    """

    def __init__(self) -> None:
        super().__init__()
        self._value = None
        self._type = ""
        self._subtype = ""

    @property
    def type(self) -> str:
        """Returns the type of the token."""
        return self._type

    @property
    def subtype(self) -> str:
        """Returns the subtype of the token."""
        return self._subtype

    @property
    def value(self) -> Any:
        """Returns the value of the token."""
        return self._value

    def __repr__(self) -> str:
        """Returns a string representation of the token."""
        return f"Token<v:{self.value} t:{self.type} s:{self.subtype} >"


class TokenBool(BaseToken):
    """
    Token class for boolean values.
    """

    def __init__(self, value: bool) -> None:
        super().__init__()
        self._value = value
        self._type = "OPERAND"
        self._subtype = "LOGICAL"


class TokenString(BaseToken):
    """
    Token class for string values.
    """

    def __init__(self, value: str) -> None:
        super().__init__()
        self._value = value
        self._type = "OPERAND"
        self._subtype = "TEXT"


class TokenNumber(BaseToken):
    """
    Token class for numeric values.
    """

    def __init__(self, value: Union[float, int]) -> None:
        super().__init__()
        self._value = value
        self._type = "OPERAND"
        self._subtype = "NUMBER"


class TokenFunc(BaseToken):
    """
    Token class for function names.
    """

    def __init__(self, func_name: str) -> None:
        super().__init__()
        self._value = func_name
        self._type = "FUNC"
        self._subtype = "OPEN"


class TokenOperator(BaseToken):
    """
    Token class for operators.
    """

    def __init__(self, operator: str) -> None:
        super().__init__()
        self._value = operator
        self._type = "OPERATOR-INFIX"
        self._subtype = ""


class TokenParen(BaseToken):
    """
    Token class for operators.
    """

    def __init__(self, subtype: str) -> None:
        super().__init__()
        self._value = "(" if subtype == "OPEN" else ")"
        self._type = "OPERATOR-INFIX"
        self._subtype = "OPEN" if subtype == "OPEN" else "CLOSE"


TokenErrorTypes = Enum(
    "TokenErrorTypes",
    [
        ("NULL", "#NULL!"),
        ("ZERO_DIV", "#DIV/0!"),
        ("VALUE", "#VALUE!"),
        ("REF", "#REF!"),
        ("NAME", "#NAME?"),
        ("NUM", "#NUM!"),
        ("NA", "#N/A"),
        ("GETTING_DATA", "#GETTING_DATA"),
    ],
)
"""
Enum class for different types of token errors.
"""


class TokenError(BaseToken):
    """
    Token class for error values.
    """

    def __init__(self, error_type: TokenErrorTypes, message: str) -> None:
        super().__init__()
        self._value = error_type.value
        self._type = "OPERAND"
        self._subtype = "ERROR"
        self._message = message

    @property
    def message(self) -> str:
        """Returns the error message."""
        return self._message


# Dictionary to store registered operations
REGISTERED_OPS: Dict[str, Callable[[Token, Token], Token]] = {}


def register_op(*names):
    """
    Decorator to register operator functions with multiple names.

    Args:
        *names: Variable length argument list of names to register the function under.

    Returns:
        The decorator function.
    """

    def decorator(func: Callable):
        for key in names:
            REGISTERED_OPS[key] = func
        return func  # Ensure the function remains usable normally

    return decorator


OpType = Callable[
    [
        Union[int, float],  # arg: a
        Union[int, float],  # arg: b
    ],
    Union[int, float],  # -> op(a,b)
]

OpRelType = Tuple[str, OpType]


def register_numeric_op(
    *pairs: OpRelType,
):
    """
    Decorator to register numeric operator functions with multiple names.

    Args:
        *pairs: Variable length argument list of tuples containing operator names and functions.

    Returns:
        The decorator function.
    """

    def decorator(
        wrap_operator: Callable[
            [OpType],
            Callable[[Token, Token], Token],
        ],
    ):
        for pair in pairs:
            key, op_func = pair
            REGISTERED_OPS[key] = wrap_operator(op_func)
        return wrap_operator  # Ensure the function remains usable normally

    return decorator



import sys

sys.path.append("src")

from mvin import BaseToken, TokenNumber, TokenString
import mvin.excel_ops as e_ops
from operator import add


class ManualToken(BaseToken):
    def __init__(self, value, type, subtype) -> None:
        super().__init__()
        self._value = value
        self._type = type
        self._subtype = subtype


def test_concat_ret_error():
    num_token = TokenNumber(2)
    assert e_ops.excel_op_concat(None, num_token).value == "#REF!"
    assert e_ops.excel_op_concat(num_token, None).value == "#REF!"


def test_eq_ret_error():
    text_token = TokenString("hi")
    assert e_ops.excel_op_eq(None, text_token).value == "#VALUE!"
    assert e_ops.excel_op_eq(text_token, None).value == "#VALUE!"


def test_neq_ok():
    text_token_a = TokenString("hi")
    text_token_b = TokenString("hi")
    assert not e_ops.excel_op_neq(text_token_a, text_token_b).value
    text_token_c = TokenString("world")
    assert e_ops.excel_op_neq(text_token_a, text_token_c).value


def test_neq_ret_error():
    text_token_a = TokenString("hi")
    assert e_ops.excel_op_neq(text_token_a, None).value == "#VALUE!"


def test_numeric_op_wrapper_ok():
    add_op = e_ops.wrap_excel_numeric_op(add)
    num_token = TokenNumber(2)
    assert add_op(num_token, num_token).value == 4


def test_numeric_op_wrapper_none_arg():
    add_op = e_ops.wrap_excel_numeric_op(add)
    num_token = TokenNumber(2)
    assert add_op(num_token, None).value == "#NUM!"  # pyright: ignore


def test_numeric_op_wrapper_error_arg():
    add_op = e_ops.wrap_excel_numeric_op(add)
    num_token = TokenNumber(2)
    error_token = ManualToken("#VALUE!", "OPERAND", "ERROR")
    assert add_op(num_token, error_token).value == "#VALUE!"  # pyright: ignore
    assert add_op(error_token, num_token).value == "#VALUE!"  # pyright: ignore


def test_numeric_op_wrapper_num_and_string():
    add_op = e_ops.wrap_excel_numeric_op(add)
    num_token = TokenNumber(2)
    text_token = TokenString("hi")
    assert add_op(num_token, text_token).value == "#NUM!"



import operator
from mvin import (
    Token,
    TokenBool,
    TokenError,
    TokenErrorTypes,
    TokenNumber,
    TokenString,
    register_numeric_op,
    register_op,
)


@register_op("&")
def excel_op_concat(a: Token, b: Token) -> Token:
    if a and b:
        if a.subtype == "ERROR":
            return a
        if b.subtype == "ERROR":
            return b
        return TokenString(f"{str(a.value)}{str(b.value)}")
    return TokenError(
        TokenErrorTypes.REF,
        "Both arguments are NoneType"
        if (a is None and b is None)
        else ("Argument a is NoneType" if a is None else "Argument b is NoneType"),
    )


@register_op("=", "==")
def excel_op_eq(a: Token, b: Token) -> Token:
    if a and b:
        if a.subtype == "ERROR":
            return a
        if b.subtype == "ERROR":
            return b

        if a.type == "OPERAND" and b.type == "OPERAND":
            return TokenBool(a.subtype == b.subtype and a.value == b.value)
    return TokenError(
        TokenErrorTypes.VALUE,
        f"Expected 2 values but, at most 1 argument was a value (a:{a} b:{b})",
    )


@register_op("<>", "!=")
def excel_op_neq(a: Token, b: Token) -> Token:
    possible_eq = excel_op_eq(a, b)
    if possible_eq and possible_eq.subtype == "LOGICAL":
        return TokenBool(not possible_eq.value)
    return possible_eq


@register_numeric_op(
    ("<", operator.lt),
    (">", operator.gt),
    ("<=", operator.le),
    (">=", operator.ge),
    ("+", operator.add),
    ("-", operator.sub),
    ("*", operator.mul),
    ("/", operator.truediv),
    ("^", operator.pow),
)
def wrap_excel_numeric_op(operator_func):
    def excel_numeric_op(a: Token, b: Token) -> Token:
        if a and b:
            if a.subtype == "ERROR":
                return a
            if b.subtype == "ERROR":
                return b
            if (a.subtype == "NUMBER" and b.subtype == "NUMBER") and (
                a.type == "OPERAND" and b.type == "OPERAND"
            ):
                return TokenNumber(operator_func(a.value, b.value))
            return TokenError(
                TokenErrorTypes.NUM,
                f"Unexpected mixed types, both should be numbers but found (a:{a} b:{b})",
            )

        return TokenError(
            TokenErrorTypes.NUM,
            "Both arguments are NoneType"
            if (a is None and b is None)
            else ("Argument a is NoneType" if a is None else "Argument b is NoneType"),
        )

    return excel_numeric_op



import sys
sys.path.append('src')

# import pytest # required for pytest.raises

from mvin.functions.excel_lib import excel_not, excel_iserror
from mvin import TokenString, TokenNumber, TokenError, TokenErrorTypes, BaseToken
from mvin.interpreter import get_interpreter

class ManualToken(BaseToken):
    def __init__(self, value, type, subtype) -> None:
        super().__init__()
        self._value = value
        self._type = type
        self._subtype = subtype

def test_excel_not_numeric_arg():
    tokens = [
        ManualToken("NOT(", "FUNC", "OPEN"),
        TokenNumber(1),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert not f({})

def test_excel_not_direct_none():
    res = excel_not(None) # pyright: ignore
    assert res is not None
    assert res != TokenErrorTypes.VALUE.value

def test_excel_iserror_false():
    tokens = [
        ManualToken("ISERROR(", "FUNC", "OPEN"),
        TokenNumber(1),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert not f({})

def test_excel_iserror_false_with_string():
    tokens = [
        ManualToken("ISERROR(", "FUNC", "OPEN"),
        TokenString("hi"),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert not f({})

def test_excel_iserror_true():
    tokens = [
        ManualToken("ISERROR(", "FUNC", "OPEN"),
        TokenError(
            TokenErrorTypes.VALUE
            , "Defined error"
        ),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({})



import sys
sys.path.append('src')

import pytest

from mvin import BaseToken, TokenError, TokenErrorTypes, TokenNumber, TokenString, TokenOperator
from mvin.interpreter import get_interpreter

class ManualToken(BaseToken):
    def __init__(self, value, type, subtype) -> None:
        super().__init__()
        self._value = value
        self._type = type
        self._subtype = subtype

def test_concat_ok():
    tokens = [TokenString("hi "), ManualToken("&", "OPERATOR-INFIX", ""), TokenString("world")]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) == "hi world"

def test_concat_left_err():
    tokens = [TokenError(
        TokenErrorTypes.REF, "Undefined message"
    ), ManualToken("&", "OPERATOR-INFIX", ""), TokenString("world")]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) == "#REF!"

def test_concat_right_err():
    tokens = [TokenString("hi "), ManualToken("&", "OPERATOR-INFIX", ""), TokenError(
        TokenErrorTypes.REF, "Undefined message"
    )]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) == "#REF!"

def test_concat_numeric_ok():
    tokens = [TokenNumber(1), ManualToken("&", "OPERATOR-INFIX", ""), TokenNumber(2)]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) == "12"

def test_add_ok():
    tokens = [TokenNumber(1), TokenOperator("+"), TokenNumber(2)]
    run = get_interpreter(tokens)
    assert run is not None
    result = run({})
    assert result == 3

def test_zerodiv_ok():
    tokens = [TokenNumber(1), ManualToken("/", "OPERATOR-INFIX", ""), TokenNumber(0)]
    run = get_interpreter(tokens)
    assert run is not None
    result = run({})
    assert result == "#DIV/0!"

def test_excel_eq_ok():
    tokens = [TokenNumber(1), ManualToken("=", "OPERATOR-INFIX", ""), TokenNumber(0)]
    run = get_interpreter(tokens)
    assert run is not None
    result = run({})
    assert result == False

def test_excel_eq_left_error():
    tokens = [TokenError(
        TokenErrorTypes.NUM, "Undefined message"
    ), ManualToken("=", "OPERATOR-INFIX", ""), TokenNumber(0)]
    run = get_interpreter(tokens)
    assert run is not None
    result = run({})
    assert result == "#NUM!"

def test_excel_eq_right_error():
    tokens = [TokenNumber(0),  ManualToken("=", "OPERATOR-INFIX", ""),TokenError(
        TokenErrorTypes.NUM, "Undefined message"
    )]
    run = get_interpreter(tokens)
    assert run is not None
    result = run({})
    assert result == "#NUM!"

def test_excel_op_raise_op_not_implemented():
    tokens = [
        TokenNumber(0),
        ManualToken("%", "OPERATOR-INFIX", "OPEN"),
        TokenNumber(1)
    ]
    f=get_interpreter(tokens)
    assert f is not None
    with pytest.raises(NotImplementedError) as exc_info:
        f({})
    assert str(exc_info.value) == "Operator '%' is not implemented"



import sys
sys.path.append('src')

import pytest # required for pytest.raises

from mvin import Token, BaseToken, TokenBool, TokenError, TokenErrorTypes, TokenNumber, TokenString
from mvin.interpreter import get_interpreter

def test_abstractToken():
    with pytest.raises(TypeError):
        base = Token() # pyright: ignore


def test_get_interpreter_ok():
    tokens = [TokenBool(True)]
    run = get_interpreter(tokens)
    assert run is not None

def test_single_bool_ok():
    tokens = [TokenBool(True)]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({})

def test_single_text_ok():
    tokens = [TokenString("hi")]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) == "hi"

def test_single_text_neq_1():
    tokens = [TokenString("hi")]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) != 1

def test_single_number():
    tokens = [TokenNumber(1)]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) == 1

def test_error_ok():
    error_token = TokenError(TokenErrorTypes.GETTING_DATA, "As expected")
    tokens = [error_token]
    run = get_interpreter(tokens)
    assert run is not None
    result = run({})
    assert result == "#GETTING_DATA"
    assert error_token.message == "As expected"

class ManualToken(BaseToken):
    def __init__(self, value, type, subtype) -> None:
        super().__init__()
        self._value = value
        self._type = type
        self._subtype = subtype

def test_range_missing_args():
    tokens = [ManualToken('F7', 'OPERAND', 'RANGE')]
    run = get_interpreter(tokens)
    assert run is not None
    with pytest.raises(KeyError):
        _ = run({})

def test_range_proper_value_transfer():
    tokens = [ManualToken('F7', 'OPERAND', 'RANGE')]
    run = get_interpreter(tokens)
    assert run is not None
    result = run({"F7": TokenString("done")})
    assert result == "done"

def test_none_as_token():
    tokens = [None]
    with pytest.raises(SyntaxError):
        get_interpreter(tokens) # pyright: ignore

def test_none_as_list():
    run_as_none = get_interpreter(None) # pyright: ignore
    assert run_as_none is None

def test_incomplete_call():
    tokens = [ManualToken("SEARCH(", "FUNC", "OPEN")]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Unmatched `(` (missing closing parenthesis)."

def test_non_existing_func():
    tokens = [ManualToken("ROBIN(", "FUNC", "OPEN"), ManualToken(")", "FUNC", "CLOSE")]
    with pytest.raises(SyntaxError):
        get_interpreter(tokens)

def test_excel_not_func_incomplete_call():
    tokens = [ManualToken("NOT(", "FUNC", "OPEN"), ManualToken(")", "FUNC", "CLOSE")]

    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)

    assert str(exc_info.value) == "Missing required argument at 0 for function `NOT(`"
    # with pytest.raises(SyntaxError) as exc_info:


    # assert str(exc_info.value) == "Function `NOT(` expects 1 arguments but got 0."

def test_excel_not_func_ok_result():
    tokens = [ManualToken("NOT(", "FUNC", "OPEN"), TokenBool(True), ManualToken(")", "FUNC", "CLOSE")]
    f = get_interpreter(tokens)
    assert f is not None
    assert not f({})


def test_double_close():
    tokens = [ManualToken("(", "PAREN", "OPEN"), ManualToken(")", "PAREN", "CLOSE"), ManualToken(")", "PAREN", "CLOSE")]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)

    assert str(exc_info.value) == "Unexpected `)` at position 2 (too many closing parentheses)."

def test_operator_after_open():
    tokens = [ManualToken("(", "PAREN", "OPEN"), ManualToken("+", "OPERATOR-INFIX", ""), ManualToken(")", "PAREN", "CLOSE")]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)

    assert str(exc_info.value) == "Unexpected operator `Token<v:+ t:OPERATOR-INFIX s: >` at position 1."

def test_missing_operator_before_par():
    tokens = [TokenNumber(2), ManualToken("(", "PAREN", "OPEN")]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)

    assert str(exc_info.value) == "Missing operator before '(' at position 1."

def test_handle_missing_last_argument():
    tokens = [
        ManualToken("NOT(", "FUNC", "OPEN"),
        TokenBool(True),
        ManualToken(",", "SEP", "ARG"),
        ManualToken(")", "FUNC", "CLOSE")
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)

    assert str(exc_info.value) == "Function `NOT(` expects 1 arguments but got 2."

def test_handle_missing_argument():
    tokens = [
        ManualToken("NOT(", "FUNC", "OPEN"),
        ManualToken(",", "SEP", "ARG"),
        ManualToken(")", "FUNC", "CLOSE")
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)

    assert str(exc_info.value) == "Function `NOT(` expects 1 arguments but got 2."

def test_handle_WSPACE():
    tokens = [
        ManualToken("", "WHITE-SPACE", "")
        , TokenBool(True)
    ]
    f = get_interpreter(tokens)
    assert f is not None

    assert f({})

def test_raise_unrecognized_token():
    tokens = [
        ManualToken("XX", "OPERATOR", ""),
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)

    assert str(exc_info.value) == "Unrecognized token `Token<v:XX t:OPERATOR s: >` at position 0."

def test_raise_unmatched_par():
    tokens = [
        ManualToken("(", "PAREN", "OPEN"), # First (
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)

    assert str(exc_info.value) == "Unmatched `(` (missing closing parenthesis)."

def test_raise_missing_operator_before_open_par():
    tokens = [
        TokenNumber(2),
        ManualToken("(", "PAREN", "OPEN"),
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)

    assert str(exc_info.value) == "Missing operator before '(' at position 1."

def test_multiple_operations():
    tokens = [
        ManualToken("(", "PAREN", "OPEN"),
        TokenNumber(2),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(0),
        ManualToken(")", "PAREN", "CLOSE"),
        ManualToken("*", "OPERATOR-INFIX", ""),
        ManualToken("(", "PAREN", "OPEN"),
        TokenNumber(10),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(2),
        ManualToken("^", "OPERATOR-INFIX", ""),
        TokenNumber(3),
        ManualToken(")", "PAREN", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == 36

def test_complex_search_call_with_warning():
    tokens = [
        TokenNumber(0),
        ManualToken("+", "OPERATOR-INFIX", ""),
        ManualToken("SEARCH(", "FUNC", "OPEN"),
        TokenNumber(1),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(1),
        ManualToken(",", "SEP", "ARG"),
        TokenNumber(5),
        ManualToken("^", "OPERATOR-INFIX", ""),
        TokenNumber(2),
        ManualToken(")", "PAREN", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == 1

def test_complex_search_call():
    tokens = [
        TokenNumber(2),
        ManualToken("+", "OPERATOR-INFIX", ""),
        ManualToken("SEARCH(", "FUNC", "OPEN"),
        TokenNumber(1),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(1),
        ManualToken(",", "SEP", "ARG"),
        TokenNumber(5),
        ManualToken("^", "OPERATOR-INFIX", ""),
        TokenNumber(2),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == 3

def test_complex_search_call_returns_not_found():
    """
    Searching 25 inside 2
    """
    tokens = [
        TokenNumber(0),
        ManualToken("+", "OPERATOR-INFIX", ""),
        ManualToken("SEARCH(", "FUNC", "OPEN"),
        TokenNumber(5),
        ManualToken("^", "OPERATOR-INFIX", ""),
        TokenNumber(2),
        ManualToken(",", "SEP", "ARG"),
        TokenNumber(1),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(1),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == "#VALUE!"


def test_incomplete_definition():
    tokens = [
        TokenNumber(0),
        TokenNumber(1)
    ]
    f = get_interpreter(tokens)
    assert f is not None
    with pytest.raises(ValueError) as exc_info:
        f({})
    assert str(exc_info.value) == 'Formula evaluation error: too many values remaining.'

def test_too_many_args():
    tokens = [
        ManualToken("SEARCH(", "FUNC", "OPEN"),
        TokenString("hi"),
        ManualToken(",", "SEP", "ARG"),
        TokenString("hi world"),
        ManualToken(",", "SEP", "ARG"),
        TokenNumber(1),
        ManualToken(",", "SEP", "ARG"),
        TokenNumber(1),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == 'Function `SEARCH(` expects 3 arguments but got 4.'

def test_double_open_par():
    tokens = [
        TokenNumber(0),
        ManualToken("+", "OPERATOR-INFIX", ""),

    ]
    f = get_interpreter(tokens)
    assert f is not None
    with pytest.raises(ValueError) as exc_info:
        f({})
    assert str(exc_info.value) == "Not enough values for operation '+'."

def test_required_missing_args():
    tokens = [
        ManualToken("SEARCH(", "FUNC", "OPEN"),
        TokenNumber(1),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(1),
        ManualToken(",", "SEP", "ARG"),
        ManualToken(",", "SEP", "ARG"),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    with pytest.raises(ValueError) as exc_info:
        f({})
    assert str(exc_info.value) == "Missing required argument at 1 for function `SEARCH(`"
